@@ -29,9 +29,36 @@ to work with.
 
 =head1 SYNOPSIS
 
-  Just a collection of functions that do not need to be in memorf if you
+  Just a collection of functions that do not need to be in memory if you
 choose one of the other methods of data storage.
 
+=head1 FORMAT
+
+The result of parsing:
+
+  <foo><head id="a">Hello <em>there</em></head><bar>Howdy<ref/></bar>do</foo>
+
+would be:
+         Tag   Content
+  ==================================================================
+  [foo, [{},
+         head, [{id => "a"},
+                0, "Hello ",
+                em, [{},
+                     0, "there"
+                    ]
+               ],
+         bar, [{},
+               0, "Howdy",
+               ref, [{}]
+              ],
+         0, "do"
+        ]
+  ]
+
+The above was copied from the XML::Parser man page.  Many thanks to
+Larry and Clark.
+
 =head1 AUTHOR
 
 By Ryan Eatmon in March 2001 for http://jabber.org/
@@ -45,7 +72,7 @@ it under the same terms as Perl itself.
 
 use vars qw($VERSION);
 
-$VERSION = "1.13";
+$VERSION = "1.15";
 
 ##############################################################################
 #
@@ -62,8 +89,7 @@ sub _handle_element {
   my ($sax, $tag, %att) = @_;
   my $sid = $sax->getSID();
 
-  $self->debug(2,"_handle_element: sid($sid) sax($sax) tag($tag) att(",%att,")")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_element: sid($sid) sax($sax) tag($tag) att(",%att,")");
 
   my @NEW;
   if($#{$self->{SIDS}->{$sid}->{tree}} < 0) {
@@ -90,8 +116,7 @@ sub _handle_cdata {
   my ($sax, $cdata) = @_;
   my $sid = $sax->getSID();
 
-  $self->debug(2,"_handle_cdata: sid($sid) sax($sax) cdata($cdata)")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_cdata: sid($sid) sax($sax) cdata($cdata)");
 
   return if ($#{$self->{SIDS}->{$sid}->{tree}} == -1);
 
@@ -99,20 +124,16 @@ sub _handle_cdata {
   $unicode->utf8($cdata);
   $cdata = $unicode->latin1;
 
-  $self->debug(2,"_handle_cdata: sax($sax) cdata($cdata)")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_cdata: sax($sax) cdata($cdata)");
 
   my $pos = $#{$self->{SIDS}->{$sid}->{tree}};
-  $self->debug(2,"_handle_cdata: pos($pos)")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_cdata: pos($pos)");
 
   if ($pos > 0 && $self->{SIDS}->{$sid}->{tree}[$pos - 1] eq "0") {
-    $self->debug(2,"_handle_cdata: append cdata")
-      unless (ref($self) eq "XML::Stream::Parser");
+    $self->debug(2,"_handle_cdata: append cdata");
     $self->{SIDS}->{$sid}->{tree}[$pos - 1] .= $cdata;
   } else {
-    $self->debug(2,"_handle_cdata: new cdata")
-      unless (ref($self) eq "XML::Stream::Parser");
+    $self->debug(2,"_handle_cdata: new cdata");
     push @{$self->{SIDS}->{$sid}->{tree}[$#{$self->{SIDS}->{$sid}->{tree}}]}, 0;
     push @{$self->{SIDS}->{$sid}->{tree}[$#{$self->{SIDS}->{$sid}->{tree}}]}, $cdata;
   }	
@@ -133,18 +154,25 @@ sub _handle_close {
   my ($sax, $tag) = @_;
   my $sid = $sax->getSID();
 
-  $self->debug(2,"_handle_close: sid($sid) sax($sax) tag($tag)")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_close: sid($sid) sax($sax) tag($tag)");
 
   my $CLOSED = pop @{$self->{SIDS}->{$sid}->{tree}};
 
-  $self->debug(2,"_handle_close: check(",$#{$self->{SIDS}->{$sid}->{tree}},")")
-    unless (ref($self) eq "XML::Stream::Parser");
+  $self->debug(2,"_handle_close: check(",$#{$self->{SIDS}->{$sid}->{tree}},")");
+
+  if ($#{$self->{SIDS}->{$sid}->{tree}} == -1) {
+    if ($self->{SIDS}->{$sid}->{rootTag} ne $tag) {
+      $self->{SIDS}->{$sid}->{streamerror} = "Root tag mis-match: <$self->{SIDS}->{$sid}->{rootTag}> ... </$tag>\n";
+    }
+    return;
+  }
 
   if($#{$self->{SIDS}->{$sid}->{tree}} < 1) {
+
     push @{$self->{SIDS}->{$sid}->{tree}}, $CLOSED;
 
-    if($self->{SIDS}->{$sid}->{tree}->[0] eq "stream:error") {
+    if(defined($self->{SIDS}->{$sid}->{tree}->[0]) &&
+       ($self->{SIDS}->{$sid}->{tree}->[0] eq "stream:error")) {
       $self->{SIDS}->{$sid}->{streamerror} =
 	$self->{SIDS}->{$sid}->{tree}[1]->[2];
     } else {
@@ -253,6 +281,7 @@ sub SetXMLData {
 #                                 the parameters
 #                     "count" - returns the number of things that match
 #                               the arguments
+#                     "tag" - returns the root tag of this tree
 #              XMLTree - pointer to XML::Parser::Tree object
 #              tag     - tag to pull data from.  If blank then the top level
 #                        tag is accessed.
@@ -269,6 +298,7 @@ sub SetXMLData {
 sub GetXMLData {
   my ($type,$XMLTree,$tag,$attrib,$value) = @_;
 
+  $tag = "" if !defined($tag);
   $attrib = "" if !defined($attrib);
   $value = "" if !defined($value);
 
@@ -323,17 +353,22 @@ sub GetXMLData {
         # mark up tags for the requested tags.
         #---------------------------------------------------------------------
 	if ($type eq "value array") {
-	  my $str = "";
-	  my $next = 0;
-	  my $index;
-	  foreach $index (1..$#{$$XMLTree[1]->[$child+1]}) {
-	    if ($next == 1) { $next = 0; next; }
-	    if ($$XMLTree[1]->[$child+1]->[$index] eq "0") {
-	      $str .= $$XMLTree[1]->[$child+1]->[$index+1];
-	      $next = 1;
+	  if ($attrib eq "") {
+	    my $str = "";
+	    my $next = 0;
+	    my $index;
+	    foreach $index (1..$#{$$XMLTree[1]->[$child+1]}) {
+	      if ($next == 1) { $next = 0; next; }
+	      if ($$XMLTree[1]->[$child+1]->[$index] eq "0") {
+		$str .= $$XMLTree[1]->[$child+1]->[$index+1];
+		$next = 1;
+	      }
 	    }
+	    push(@array,$str);
+	  } else {
+	    push(@array,$$XMLTree[1]->[$child+1]->[0]->{$attrib})
+	      if (exists $$XMLTree[1]->[$child+1]->[0]->{$attrib});
 	  }
-	  push(@array,$str);
 	}
         #---------------------------------------------------------------------
 	# Return a pointer to a new XML::Parser::Tree object that has the
@@ -395,14 +430,14 @@ sub GetXMLData {
       return $count;
     }
   } else {
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     # This is the root tag, so handle things a level up.
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
 
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     # Return the raw CDATA value without mark ups, or the value of the
     # requested attribute.
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     if ($type eq "value") {
       if ($attrib eq "") {
 	my $str = "";
@@ -420,28 +455,34 @@ sub GetXMLData {
       return $$XMLTree[1]->[0]->{$attrib}
         if (exists $$XMLTree[1]->[0]->{$attrib});
     }
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     # Return a pointer to a new XML::Parser::Tree object that has the
     # requested tag as the root tag.
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     if ($type eq "tree") {
       my @tree =  @{$$XMLTree};
       return @tree;
     }
 
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     # Return the 1 if the specified attribute exists in the root tag.
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     if ($type eq "existence") {
       return 1 if (($attrib ne "") && (exists($$XMLTree[1]->[0]->{$attrib})));
     }
 
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     # Return the attribute hash that matches this tag
-    #---------------------------------------------------------------------
+    #-------------------------------------------------------------------------
     if ($type eq "attribs") {
       return %{$$XMLTree[1]->[0]};
     }
+    #-------------------------------------------------------------------------
+    # Return the tag of this node
+    #-------------------------------------------------------------------------
+    if ($type eq "tag") {
+      return $$XMLTree[0];
+    }
   }
   #---------------------------------------------------------------------------
   # Return 0 if this was a request for existence, or "" if a request for
